# Tuple is (asm, marker, color)
OPS= {
  0x00: ("NOP",                   "  ", 0),
  0x01: ("MOV, RI, 0",            "I0", 0),
  0x02: ("MOV, RI, 1",            "I1", 0),
  0x03: ("MOV, RI, 2",            "I2", 0),
  0x04: ("MOV, RI, 3",            "I3", 0),
  0x05: ("MOV, RI, 4",            "I4", 0),
  0x06: ("MOV, RI, 5",            "I5", 0),
  0x07: ("MOV, RI, 6",            "I6", 0),
  0x08: ("MOV, RI, 7",            "I7", 0),
  0x09: ("MOV R[RI], R0",         "M0", 0),
  0x0A: ("MOV R[RI], R1",         "M1", 0),
  0x0B: ("MOV R[RI], R2",         "M2", 0),
  0x0C: ("MOV R[RI], R3",         "M3", 0),
  0x0D: ("MOV R[RI], R4",         "M4", 0),
  0x0E: ("MOV R[RI], R5",         "M5", 0),
  0x0F: ("MOV R[RI], R6",         "M6", 0),
  0x10: ("MOV R[RI], R7",         "M7", 0),
  0x11: ("INC R[RI]",             "++", 0),
  0x12: ("DEC R[RI]",             "--", 0),
  0x13: ("ADD R[RI], R6, R7",     "AD", 0),
  0x14: ("SUB R[RI], R6, R7",     "SB", 0),
  0x15: ("MUL R[RI], R6, R7",     "ML", 0),
  0x16: ("DIV R[RI], R6, R7",     "//", 0),
  0x17: ("MOD R[RI], R6, R7",     "%%", 0),
  0x18: ("INV R[RI]",             "~~", 0),
  0x19: ("NEG R[RI]",             "NG", 0),
  0x1A: ("AND R[RI], R6, R7",     "&&", 0),
  0x1B: ("OR  R[RI], R6, R7",     "||", 0),
  0x1C: ("XOR R[RI], R6, R7",     "^^", 0),
  0x1D: ("EQ, R[RI], R6, R7",     "==", 0),
  0x1E: ("LT, R[RI], R6, R7",     "<<", 0),
  0x1F: ("LINK",                  "🡷🡷", 95),
  0x20: ("SKIP?",                 "??", 94),
  0x21: ("LOAD  R[RI], [R0.R1]",  "LD", 36),
  0x22: ("STORE R[RI], [R0.R1]",  "ST", 36),
  0x23: ("RETURN",                "🞖🞖", 33),
  0x24: ("GETC R[RI]",            "II", 32),
  0x25: ("PUTC R[RI]",            "OO", 33),
  0x26: ("SET, DIR, LEFT",        "🡰🡰", 0),
  0x27: ("SET, DIR, DOWN",        "🡳🡳", 0),
  0x28: ("SET, DIR, RIGHT",       "🡲🡲", 0),
  0x29: ("SET, DIR, UP",          "🡱🡱", 0),
}

def disasm(ops):
  if type(ops) != list:
    ops = ops.flatten()

  for o in ops:
    code,marker,color = ("INVALID", "XX", 31)
    if o in OPS:
      code,marker,color = OPS[o]

    rendered = f"\033[12;{color}m{code}\033[0m"
    print(rendered)

def reasm(markers):
  mapping = { marker : opcode for opcode, (asm,marker,color) in OPS.items() }

  ret = []
  for m in markers:
    if m not in mapping:
      print(f"Warning: Unknown instruction mnemonic {m}. NOPing out.")
      m = "  "    # NOP

    ret.append(mapping[m])

  return ret

# Pass in all the bytes from vm->memory()
def render_field(memory_dump, pc_row=None, pc_col=None):

  maze = []
  for row in range(70):
    r = []
    for col in range(70):
      op = memory_dump[(256 * col) + row]
      r.append(op)
    maze.append(r)


  # X axis (these are code banks)
  X_AXIS="   "
  for col in range(len(maze[0])):
    X_AXIS += f" {col:02x}"

  print(X_AXIS)
  rowno = 0


  for y in range(len(maze)):
    r = maze[y]
    txt = f" {rowno:02x}"
    for x in range(len(r)):
      opcode = r[x]

      code,marker,color = "INVALID", "XX", 31
      if opcode in OPS:
        code,marker,color = OPS[opcode]

      #if opcode not in [ 0, 0x26, 0x27, 0x28, 0x29 ] or marker == "":
      if marker == "":
        marker = f"{opcode:02x}"
      else:
        marker = marker[:2]

      if (pc_row == y and pc_col == x):
        marker = f"\033[3m{marker}\033[0m"
      elif color != 0:
        marker = f"\033[{color}m{marker}\033[0m"

      txt += f" {marker}"


    txt += f" {rowno:02x}"
    rowno += 1
    print(txt)
  print(X_AXIS)

def parse_regs(vm_dump):

  ret = {}

  pc = vm_dump[0x10008:0x1000A]
  ret["pch"] = pc[0]
  ret["pcl"] = pc[1]

  regs = vm_dump[:8]
  ret["r"] = regs

  return ret

# Pass in the entirety of the VM object
def render_regs(vm_dump):

  regs = parse_regs(vm_dump)

  text = ""

  text += f" PC={regs['pch']:02x},{regs['pcl']:02x}"

  for i in range(8):
    text += f"  R{i}={regs['r'][i]:02x}"

  text += f" PTR=0x{regs['r'][0]:02x}{regs['r'][1]:02x}"
  print(text)
  return regs

def render_full_state(vm_dump):
  regs = render_regs(vm_dump)

  memory_dump = vm_dump[8:8+65536]
  render_field(memory_dump, regs['pcl'], regs['pch'])

