#!/usr/bin/env python3

import argparse
from pwn import *

import dicevm


binary = "./lost_in_your_eyes"
server = "dicec.tf"
port   = 31578

def banner(text="Hack Successful."):
  try:
    from pyfiglet import Figlet
    f = Figlet(font='slant')
    text = f.renderText(text)
  except:
    pass

  try:
    from lolpython import lol_py
    lol_py(text)
  except:
    print(text)

def connect():
  global server
  global port
  if (conf.remote):
    target = remote(server, port)
  else:
    target = process([binary])
  return target

try:
  elf = ELF(binary)
  #libc = ELF("lib.so")
  context.binary = elf
except:
  print(f"Unable to load binary {binary}, falling back to default settings")
  context.arch = 'amd64'
  context.os = 'linux'
  context.endian = 'little'
  context.word_size = 64

context.aslr = False


parser = argparse.ArgumentParser(description="Hack the planet")

parser.add_argument('--debug', '-d', action='store_true',
    help="Show more debugging info")

parser.add_argument('--remote', '-r', action='store_true',
    help="Perform attack on remote host")

parser.add_argument('--pause', '-p', action='store_true',
    help="Pause after spawning local process")


conf = parser.parse_args()


# Take the big string list below and transform it into the byte stream
# that the VM expects as input. You should eventually see your program
# in the death box!
def program_from_2d_text(source):
  if len(source) != 47:
    print(f"Expected 47 source lines, found {len(source)}")
  if any([ len(l) != (35*2+33) for l in source]):
    print(f"At least one row is of the wrong length.")

  # Each mnemonic is two characters, and then they are separated by a single space
  source_split = [ [ line[s:s+2] for s in range(0,len(line),3) ] for line in source ]

  # Cut the program up into single mnemonics, still organized by line
  if any([ len(l) != 35 for l in source_split]):
    print("At least one row has the wrong number of mnenomics, somehow.")


  # Reorder the mnemonics in the order we have to load them, i.e. top-to-bottom and then left-to-right
  program_text = []
  for col in range(len(source_split[0])):
    for row in range(len(source_split)):
      program_text.append(source_split[row][col])

  program = dicevm.reasm(program_text)
  return program

# the program box is 35 ops wide and 47 ops tall.
# Entry is the bottom-left corner (coming in from the left)
# Exit is the top-left corner (exiting left)
# Here's an alignment and OP chart. Two spaces is a NOP.
# I0 I1 I2 I3 I4 I5 I6 I7 M0 M1 M2 M3 M4 M5 M6 M7 ++ -- AD SB ML // %% ~~ NG && || ^^ == << ?? LD 🡰🡰 🡳🡳
# 🡲🡲 🡱🡱
source = [
 "🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰",
 "      🡲🡲 🡳🡳                                           ++                                           🡱🡱",
 "🡲🡲 🡳🡳 ?? I4                                           I4                                           🡱🡱",
 "      ~~ ++                                        🡱🡱 🡱🡱 ?? ~~ == M0 I7 -- -- I6    🡰🡰             🡱🡱",
 "      == ++                                                                      🡳🡳 🡰🡰             🡱🡱",
 "      M0 --                                                                      I4 ??             🡱🡱",
 "      I7          ++                                                             ++ ~~             🡱🡱",
 "      --          I4                                                                ==             🡱🡱",
 "      --       🡱🡱 🡰🡰          🡳🡳 ?? ~~ == M0 I7 SB I6 M1 I7    🡳🡳 ??    ~~    🡰🡰    M0             🡱🡱",
 "         🡳🡳 🡰🡰    ??          🡲🡲 I4 ++          🡳🡳    ++ ++ I4 🡰🡰             ==    I7             🡱🡱",
 "         I4 ??    🡱🡱    == M0 I7 -- I6 🡳🡳 ??       ~~ == M0 I7 -- --    🡰🡰    M0    --             🡱🡱",
 "      SB       🡱🡱R🡰🡰E🡰🡰T🡰🡰U🡰🡰R🡰🡰N🡰🡰!🡰🡰X  X🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰 🡰🡰    I6    I7    --             🡱🡱",
 "      I6 --                            I4    🡳🡳 🡰🡰             🡳🡳   X  X🡰🡰    --    --             🡱🡱",
 "      M5 --          🡳🡳 🡰🡰       🡳🡳 🡰🡰 ++    I4 ?? 🡳🡳 🡰🡰    🡰🡰          ??    --    --             🡱🡱",
 "      I7       🡳🡳 🡰🡰    ??       I4 ??       ++ ~~    ??    ?? 🡳🡳 🡰🡰    ~~    I6    I6             🡱🡱",
 "      🡲🡲 🡳🡳    I4 ?? I4 ~~ 🡳🡳 🡰🡰 ++ ~~ 🡳🡳 🡰🡰 ++ == I4       ~~ I4 ??    == 🡳🡳 🡰🡰 🡳🡳 🡰🡰 🡳🡳 🡰🡰       🡱🡱",
 "      ?? I4    ++    ++ == I4 ?? ++ == I4 ?? ++ M0 ++       == ++ ~~    M0 I4 ?? I4 ?? I4 ??       🡱🡱",
 "      ~~ ++    ++    ++ M0 ++ ~~ ++ M0 ++ ~~ ++ I7 ++       M0 ++ ==    I7 ++ ~~ ++ ~~ ++ ~~       🡱🡱",
 "      == ++ ~~ ++ ~~ ++ I7 ++ ==    I7    ==    -- ++ ~~    I7 ++ M0    -- ++ ==    ==    ==       🡱🡱",
 "++    M0 ++ == ++ ==    SB ++ M0    --    M0    -- ++ ==    -- ++ I7    -- ++ M0    M0    M0       🡱🡱",
 "++    I7    M0    M0    I6 ++ I7    I6    I7    I6    M0    --    --    --    I7    I7    I7       🡱🡱",
 "++    SB 🡲🡲X  X🡲🡲X  X🡲🡲X  X🡲🡲X  X🡲🡲X  X🡲🡲X  X🡲🡲X  X🡲🡲X  X  X  X🡲🡲X  X🡲🡲X  X🡲🡲X  X🡲🡲X  X🡲🡲X  X  X   🡱🡱",
 "++    I6 ++ I7    I7    M5    --          -- 🡳🡳 🡰🡰    I7    --    -- ++ -- ++ --    --    --       🡱🡱",
 "++    M5 ++ -- ++ --    I7    -- 🡳🡳 🡰🡰 ++ --    ??    --    --    -- I4 I6 ++ --    -- ++ --       🡱🡱",
 "++    I7 I4 I6 I4 I6 🡳🡳 🡰🡰 ++ I6 I4 ?? ++ -- I4 ~~ ++ I6    I6    I6 🡱🡱 🡰🡰 I4 I6    I6 ++ --       🡱🡱",
 "++                      ?? I4    ++ ~~ I4 I6 ++ == I4          🡳🡳 🡰🡰    ?? 🡱🡱 🡰🡰 🡰🡰    I4 I6       🡱🡱",
 "M0    🡲🡲 🡱🡱 🡰🡰 🡱🡱 🡰🡰 I4 ~~ 🡱🡱 🡰🡰 🡰🡰 == 🡱🡱 🡰🡰    M0 🡱🡱 🡰🡰 🡰🡰 🡰🡰 I4 ??    ~~    ?? 🡱🡱 🡰🡰 🡱🡱 🡰🡰       🡱🡱",
 "I5    ??    ??    ?? ++ ==    ??    M0    ??    I7    ??    ?? ++ ~~    ==    ~~    ??    ??       🡱🡱",
 "      ~~    ~~    ~~    M0    ~~    I7    ~~          ~~    ~~ ++ ==    M0    ==    ~~    ~~       🡱🡱",
 "--    ==    ==    ==    I7    ==    --    ==          ==    ==    M0    I7    M0    ==    ==       🡱🡱",
 "M0 I6 M0    M0    M0    --    M0    --    M0    --    M0    M0    I7    --    I7    M0    M0       🡱🡱",
 "I4 M1 I7    I7    I7    --    I7    SB    I7    SB    I7    I7    --    SB    --    I7    I7       🡱🡱",
 "      --       🡱🡱    🡰🡰 SB          SB 🡱🡱    🡰🡰 SB 🡱🡱          🡰🡰 SB    SB    SB    --             🡱🡱",
 "^^ I1 --    --    --    I6    --    I6    --    I6    --    --    I6    I6    I6    --    --       🡱🡱",
 "I0 M0                   M5    --    M5          M5          --    M5    M1    M5    --             🡱🡱",
 "M6 ++                   I7          I7          I7                I7    I7    I7    --             🡱🡱",
 "I7 ++ M2    M2    M2    M2    M2    M2    M2    M2    M2    M2    M2    M2    M2    M2    M2       🡱🡱",  # Top of column match
 "   ++ I6    I6    I6    I6    I6    I6    I6    I6    I6    I6    I6    I6    I6    I6    I6       🡱🡱",
 "M7 ++ 🡲🡲 🡳🡳 🡲🡲 🡳🡳 🡲🡲 🡳🡳 🡲🡲 🡳🡳 🡲🡲 🡳🡳 🡲🡲 🡳🡳 🡲🡲 🡳🡳 🡲🡲 🡳🡳 🡲🡲 🡳🡳 🡲🡲 🡳🡳 🡲🡲 🡳🡳 🡲🡲 🡳🡳 🡲🡲 🡳🡳 🡲🡲 🡳🡳 🡲🡲 🡳🡳    🡱🡱",
 "I2 ++ ??    ??    ??    ??    ??    ??    ??    ??    ??    ??    ??    ??    ??    ??    ??       🡱🡱",
 "      ==    ==    ==    ==    ==    ==    ==    ==    ==    ==    ==    ==    ==    ==    ==       🡱🡱",
 "M6    M0    M0    M0    M0    M0    M0    M0    M0    M0    M0    M0    M0    M0    M0    M0       🡱🡱",
 "I1    I7    I7    I7    I7    I7    I7    I7    I7    I7    I7    I7    I7    I7    I7    I7       🡱🡱",
 "      --    --    --    --    --    --    --    --    --    --    --    --    --    --    --       🡱🡱",
 "                  --                                                          --          --       🡱🡱",
 "      I6    I6    I6    I6    I6    I6    I6    I6    I6    I6    I6    I6    I6    I6    I6       🡱🡱",
 "🡱🡱 🡲🡲 🡱🡱 🡲🡲 🡱🡱 🡲🡲 🡱🡱 🡲🡲 🡱🡱 🡲🡲 🡱🡱 🡲🡲 🡱🡱 🡲🡲 🡱🡱 🡲🡲 🡱🡱 🡲🡲 🡱🡱 🡲🡲 🡱🡱 🡲🡲 🡱🡱 🡲🡲 🡱🡱 🡲🡲 🡱🡱 🡲🡲 🡱🡱 🡲🡲 🡱🡱 🡲🡲 🡲🡲 🡱🡱"
]
# I0 I1 I2 I3 I4 I5 I6 I7 M0 M1 M2 M3 M4 M5 M6 M7 ++ -- AD SB ML // %% ~~ NG && || ^^ == << ??
# LD 🡰🡰 🡳🡳 🡲🡲 🡱🡱

#       C1    C2    C4    C5    C6    C7    C8    C9    C10   C11  C12    C13   C15   C16   C18

# Top and right edges are return sumps, they all lead to the exit
# Left edge up and then down is our preamble, to set up whatever needs setting up
#
# For now, it stores Column in R1 and row in R2, and zeroes out R0
# It also sets R4 to 255, to use as a default case
# R5 is set to 7, to use for fast decrementation
# Likewise, R1 is set to 5.

# Once the main program starts, there are a series of columns of code. Each column corresponds to one
# column in our lookup table. The code columns are in order, so that 0 comes first.
#
# The beginning of each code column decrements R6 so that a check against 0 is proper.
# If R6 is then 0, we have a column match and continue upward.
# This starts a row match, in a similar fashion.
# We check all the rows we are interested in for this column. If there's a hit,
# we need to increment R4 from its starting value of 255 up to the proper move direction.
# The space between each code column is used to set up R4 and navigate to a return somehow.

# The space between operations is deliberately ignored by the parser, so we use them for documentation
# For example, the row of
#   X  X🡲🡲X  X🡲🡲X  X🡲🡲X  X🡲🡲X  X🡲🡲X  X🡲🡲X  X🡲🡲X  X  X  X🡲🡲X 
# across the center of the program is a highway to exit. used by many of the towers.


program = program_from_2d_text(source)

target = connect()

# Fill up the death box with our program, short one byte
target.send(bytes(program[:len(program)-1]))

# Run this script with --pause and then connect GDB to the lost_in_your_eyes process
# Press ENTER to finish feeding input so the VM continues executing.
if conf.pause: pause()

target.send(bytes(program[-1:]))

# There's a little
target.send(bytes([0x00]*4))

banner()

target.interactive()


